from collections import ChainMap, Mapping
from functools import partial
from os import PathLike
from pathlib import Path
from typing import AnyStr, Dict, NoReturn, Union, Optional

import yaml

from ConfigFramework.abstract.abc_loader import AbstractConfigLoader

try:
    from yaml import CLoader as Loader, CDumper as Dumper

except ImportError:
    from yaml import Loader, Dumper


class YAMLLoader(AbstractConfigLoader):
    dumper = partial(yaml.dump, Dumper=Dumper)
    loader = partial(yaml.load, Loader=Loader)

    def __init__(self, data: Union[Dict, ChainMap, Mapping], defaults: Dict, config_path: Path):
        super().__init__(data, defaults)
        self.config_path = config_path

    @classmethod
    def load(cls, config_path: Union[AnyStr, Path, PathLike], defaults: Optional[Dict] = None):
        config_path = Path(config_path)

        if not config_path.is_file():
            raise ValueError(f"Invalid file path: {config_path}")

        with open(config_path, encoding='utf8') as yaml_f:
            return cls(cls.loader(yaml_f), defaults, config_path)

    def dump(self, include_defaults: bool = False) -> NoReturn:
        to_dump = self.data

        if include_defaults:
            to_dump = dict(self.lookup_data)

        with open(self.config_path, 'w', encoding="utf8") as yaml_f:
            self.dumper(to_dump, yaml_f)

    def __str__(self):
        return f"{self.__class__.__name__}: {self.config_path}"




from .composite_loader import CompositeLoader
from .env_loader import EnvLoader
from .json_loader import JsonLoader
from .json_string_loader import JsonStringLoader
from .yaml_loader import YAMLLoader



from os import environ
from typing import Dict, NoReturn, Optional

from ConfigFramework.abstract.abc_loader import AbstractConfigLoader


class EnvLoader(AbstractConfigLoader):
    @classmethod
    def load(cls, defaults: Optional[Dict] = None):
        return cls(data=environ, defaults=defaults)

    def dump(self, include_defaults: bool = False) -> NoReturn:
        """
        You can not dump values to environ yet because I don't know a good way to do that or think it's needed.
        Left only for compatibility purpose.

        :param include_defaults: doesn't affect anything as stated before.
        :return:
        """
        pass



from __future__ import annotations
from functools import wraps
from typing import Any, AnyStr, Callable, Hashable, NoReturn, Optional, TYPE_CHECKING, Type
from ConfigFramework.loaders.composite_loader import CompositeLoader

if TYPE_CHECKING:
    from .abc_loader import AbstractConfigLoader
    from ConfigFramework.dump_caster import DumpCaster


class AbstractConfigVar:
    """
    Abstract config variable with descriptors interface which is a base type for any of your variables classes.

    """
    def __init__(
        self, key: [Hashable, AnyStr], loader: AbstractConfigLoader, *,
        typehint: Optional[Type] = Any,
        caster: Optional[Callable] = None, dump_caster: Optional[Callable, DumpCaster] = None,
        validator: Optional[Callable] = None, default: Optional[Any] = None, constant: bool = False
    ):
        """
        Initializes variable for specified first_loader and key.

        :param key: Any hashable or a string. Strings can be written as paths in case you need a variable.
         that underlies other mappings. Example of how to get such vars: `config_root/database/database_ip`.
        :param loader: A first_loader that will be looked up to get vars config_var or to update values.
        :param typehint: Typehint for __value field, that by default being returned.
        :param caster: Callable that should return variable casted to specific type (in case you need custom types).
        :param dump_caster: Callable that being called when config being dumped. Also can be instance of
         ConfigFramework.dump_caster.DumpCaster
        :param validator: Callable that validates config_var or defaults in case the original config_var is invalid.
        :param default: Default config_var that will be set, if config_var is invalid.
        :param constant: Sets if variable config_var can be set in runtime
        """
        self.key = key
        self.is_constant = constant
        self.loader: AbstractConfigLoader = loader
        self.__post_init = False

        # Redefining functions that we will need if they are provided
        for redefine_func_name, func in zip(
            ("caster", "dump_caster", "validate"),
            (caster, dump_caster, validator)
        ):
            if callable(func):
                setattr(self, redefine_func_name, func)

        if default is not None:
            self.__value: typehint = self.caster(loader.get(key, default))
            default = caster(default)

        else:
            self.__value: typehint = self.caster(loader[key])

        # Validation of config_var and defaults
        bool_casted_validator: Callable = self._validate_to_bool(self.validate)

        if not bool_casted_validator(self.__value):
            if (default is not None) and bool_casted_validator(default):
                self.__value: typehint = default

            elif self._validate_value_in_defaults(self.validate, self.caster, loader, key):
                key = loader.key_to_path_cast(key)
                var_key = loader.key_to_path_cast(key)

                self.value: typehint = loader.get_to_variable_root(key, lookup_at=loader.defaults)[var_key]

            else:
                raise ValueError(f"Invalid config_var for {self} in {self.loader} and default values not found")

        self.__post_init = True

    def caster(self, value: Any) -> Any:
        """
        Callable that should return variable casted to specific type (in case you need custom types).

        :param value: config_var to be casted
        :return: value casted to whatever type you need.
        """
        return value

    def dump_caster(self, config_var: AbstractConfigVar) -> Any:
        """
        Callable that being called when config being dumped.
        Nothing being passed as attribute since it should be executed after assigning the config_var to ConfigVar.

        Through self we can obtain access to useful stuff and also use DumpCaster, which allows us to assign
        specific caster for exact ConfigLoader
        :return: value casted to whatever type you need to save value.
        """
        return config_var.__value

    def validate(self, value: Any) -> bool:  # noqa
        """
        Callable that validates config_var or defaults in case the original config_var is invalid.
        :param value: config_var to be validated
        :return: bool value representing if variable has a valid value.
        """
        return True

    def _loader_type(self) -> Type[AbstractConfigLoader]:
        """
        Internal function that helps us to find from what exact first_loader we got var from.
        :return: loader class.
        """
        if not isinstance(self.loader, CompositeLoader):
            return type(self.loader)

        else:
            self.loader: CompositeLoader
            for loader in self.loader.loaders:
                try:
                    loader[self.key]

                except KeyError:
                    continue

                else:
                    return type(loader)

    @staticmethod
    def _validate_to_bool(func):
        """
        Wrapper that returns a bool config_var representing if check failed or raised ValueError.

        :param func: function that is wrapped for specific validator. Will be used in checking if value is correct
        to apply defaults.
        :return: callable.
        """
        @wraps(func)
        def execute(*args, **kwargs):
            try:
                return func(*args, **kwargs)

            except ValueError:
                return False

        return execute

    @staticmethod
    def _validate_value_in_defaults(validator: Callable, caster: Callable, loader: AbstractConfigLoader, key) -> bool:
        if not hasattr(loader, 'defaults'):
            return False

        bool_casted_validator: Callable = AbstractConfigVar._validate_to_bool(validator)
        *keys, variable_key = loader.key_to_path_cast(key)

        try:
            variable_root = loader.get_to_variable_root(keys, lookup_at=loader.defaults)
            variable = variable_root[variable_key]

        except KeyError:
            return False

        return bool_casted_validator(caster(variable))

    @property
    def value(self) -> Any:
        """
        Gives access to casted value.

        :return: any value from config, that been casted and validated.
        """
        return self.__value

    @value.setter
    def value(self, value: Any) -> NoReturn:
        """
        Sets a new value to config variable and updates its value in loader with casting it to needed type.

        :param value:
        :return:
        """
        if self.is_constant and self.__post_init:
            raise NotImplementedError("Constants can not be assigned in runtime")

        if not self.validate(value):
            raise ValueError(f"Invalid config_var to be set for property {self}")

        self.__value = value
        self.loader[self.key] = self.dump_caster(self)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f"{self.key} in {self.loader} = {self.__value}"


__all__ = ["AbstractConfigVar"]



from __future__ import annotations

from abc import ABC, abstractmethod
from collections import ChainMap, Mapping
from functools import lru_cache
from pathlib import Path
from time import time
from typing import Any, AnyStr, Dict, Hashable, NoReturn, Optional, Tuple, Union


class AbstractConfigLoader(ABC, Mapping):
    """
    Base class for any of your loaders. Initialization can be made through load function, or, if author wants so,
    through `__init__` function.
    """

    def __init__(
        self, data: Union[Dict, ChainMap, Mapping], defaults: Dict, include_defaults_to_dumps: Optional[bool] = None,
        *args, **kwargs
    ):
        """
        Creates new config first_loader with your variables.

        :param data: data from first_loader.
        :param defaults: default variables that will be used if not found in first_loader.
        :param include_defaults_to_dumps: represents if default values should be dumped to that first_loader.
        :param args: arguments that can be used for your custom loaders.
        :param kwargs: keyword arguments that can be used for your custom loaders.
        """
        self.__created_at = str(time())
        self.data = data
        self.include_defaults: bool = include_defaults_to_dumps
        self.defaults = defaults

        if isinstance(defaults, dict):
            self.lookup_data = ChainMap(self.data, self.defaults)

        else:
            self.lookup_data = self.data

    @classmethod
    @abstractmethod
    def load(cls, *args, **kwargs):
        """
        Initializes creation of new ConfigLoader.
        You must provide your arguments that are needed for your first_loader.

        :param args: arguments that can be used for your custom loaders.
        :param kwargs: keyword arguments that can be used for your custom loaders.
        """
        pass

    @abstractmethod
    def dump(self, include_defaults: bool = False) -> NoReturn:
        """
        Dumps updated variables.

        :param include_defaults: specifies if you want to have default variables to be dumped.
        :return:
        """
        pass

    def dump_to(self, other_loader: AbstractConfigLoader, include_defaults: bool = False) -> NoReturn:
        """
        Dumps variables to other first_loader that been already initialized.

        :param other_loader: other first_loader that already initialized and where you want to dump stuff.
        :param include_defaults: include_defaults_to_dumps: specifies if you want to have default variables to be.
        :return: nothing.
        """
        if isinstance(other_loader, AbstractConfigLoader):
            if include_defaults:
                other_loader.lookup_data = self.lookup_data

            else:
                other_loader.lookup_data = self.data

        else:
            raise ValueError(
                f"{self} got invalid first_loader to dump variables to. Got type: {type(other_loader)}"
            )

        other_loader.dump(include_defaults)

    def get(self, key: Union[Hashable, AnyStr], default=None) -> Any:
        """
        Returns an item under specified key or key as path and if
        it didn't find it - returns default variable.

        :param key: a key that points at where you want to grab variable.
        :param default: default config_var.
        :return: anything lying there.
        """
        casted_key = self.key_to_path_cast(key)
        val_root = self.get_to_variable_root(casted_key)

        return val_root.get(casted_key[-1], default)

    @staticmethod
    @lru_cache(maxsize=None)
    def key_to_path_cast(key: Union[AnyStr, Hashable]) -> Union[Tuple[AnyStr, ...], Tuple[Hashable, ...]]:
        """
        Casts a key to tuple of keys, that should be applied one by one to get to where variable lies.
        Path pointing example: config_root/database/database_ip

        :param key: string with path, pointing at out variable or hashable.
        :return: tuple of sub keys.
        """
        if isinstance(key, str) and ("/" in key):
            return Path(key).parts

        return key,

    def get_to_variable_root(
        self, keys: Union[Tuple[AnyStr, ...], Tuple[Hashable, ...]], lookup_at: Optional[Dict] = None
    ) -> Dict:
        """
        Returns an dictionary with our variable.

        :param keys: keys tuple we apply to get to root of variable.
        :param lookup_at: the location we're looking at. In case we need to lookup at specific part of our first_loader.
        :return: a dictionary which layer should contain our variable.
        """
        root = lookup_at or self.lookup_data
        # Last part of key must be the variable
        for key in keys[:-1]:
            root = root[key]

        return root

    def __getitem__(self, key: Union[Hashable, AnyStr]) -> Any:
        """
        Returns an item under specified key or key as path and if
        it didn't find it - raises KeyError.

        :param key: variable key.
        :return: value.
        """
        casted_key = self.key_to_path_cast(key)
        val_root = self.get_to_variable_root(casted_key)

        return val_root[casted_key[-1]]

    def __setitem__(self, key: Union[Hashable, AnyStr], value: Any) -> NoReturn:
        """
        Sets variable config_var inside of your first_loader.

        :param key: a key that points at what variable you want to set config_var to.
        :param value: a new config_var for variable.
        :return: nothing.
        """

        casted_key = self.key_to_path_cast(key)
        val_root = self.get_to_variable_root(casted_key)

        if casted_key[-1] not in val_root:
            raise KeyError(f"No key {casted_key[-1]} in {self}")

        val_root[casted_key[-1]] = value

    def __hash__(self):
        return hash(self.__class__.__name__ + self.__created_at)

    def __repr__(self):
        return f"Loader: {self.__class__.__name__}"

    def __str__(self):
        return self.__class__.__name__

    def __len__(self) -> int:
        return len(self.lookup_data)

    def __iter__(self):
        return iter(self.lookup_data)


__all__ = ["AbstractConfigLoader"]



import json
from collections import ChainMap, Mapping
from os import PathLike
from functools import partial
from pathlib import Path
from typing import AnyStr, Dict, NoReturn, Union, Optional

from ConfigFramework.abstract.abc_loader import AbstractConfigLoader


class JsonLoader(AbstractConfigLoader):
    json_loader = partial(json.load)
    json_dumper = partial(json.dump, ensure_ascii=False, check_circular=True, indent=2)

    def __init__(self, data: Union[Dict, ChainMap, Mapping], defaults: Dict, config_path: Path):
        super().__init__(data, defaults)
        self.config_path = config_path

    @classmethod
    def load(cls, config_path: Union[AnyStr, Path, PathLike], defaults: Optional[Dict] = None, **json_kwargs):
        config_path = Path(config_path)

        if not config_path.is_file():
            raise ValueError(f"Invalid file path: {config_path}")

        with config_path.open(mode='r', encoding='utf-8') as json_f:
            data = cls.json_loader(json_f, **json_kwargs)

        return cls(data, defaults, config_path=config_path)

    def dump(self, include_defaults: bool = False, **json_kwargs) -> NoReturn:
        to_dump = self.data

        if include_defaults:
            to_dump = dict(self.lookup_data)

        with open(self.config_path, 'w', encoding="utf8") as json_f:
            self.json_dumper(to_dump, json_f, **json_kwargs)

    def __str__(self):
        return f"{self.__class__.__name__}: {self.config_path}"



import json
from collections import ChainMap, Mapping
from functools import partial
from typing import AnyStr, Dict, Optional, Union

from ConfigFramework.abstract.abc_loader import AbstractConfigLoader


class JsonStringLoader(AbstractConfigLoader):
    json_serialized_loader = json.loads
    json_serialized_dumper = partial(json.dump, ensure_ascii=False, check_circular=True, indent=2)

    def __init__(self, data: Union[Dict, ChainMap, Mapping], defaults: Dict):
        super().__init__(data, defaults)

    @classmethod
    def load(cls, config: AnyStr, defaults: Optional[Dict] = None, **json_kwargs):
        data = cls.json_serialized_loader(config, **json_kwargs)

        return cls(data, defaults)

    def dump(self, include_defaults: bool = False, **json_kwargs) -> AnyStr:
        to_dump = self.data

        if include_defaults:
            to_dump = dict(self.lookup_data)

        return self.__class__.json_serialized_dumper(to_dump, ensure_ascii=False, check_circular=True, **json_kwargs)



from __future__ import annotations

from typing import Any, AnyStr, Callable, Hashable, Optional, Tuple, Type

from ConfigFramework.abstract.abc_loader import AbstractConfigLoader
from ConfigFramework.abstract.abc_variable import AbstractConfigVar
from .dump_caster import DumpCaster


class ConfigVar(AbstractConfigVar):
    # Yes, its literally the same thing
    """
    Represents any config variable that you can customize to your needs by adding some code.

    """

    def __init__(
        self, key: [Hashable, AnyStr], loader: AbstractConfigLoader, *,
        typehint: Optional[Type] = Any,
        caster: Optional[Callable] = None, dump_caster: Optional[Callable, DumpCaster] = None,
        validator: Optional[Callable] = None, default: Optional[Any] = None, constant: bool = False
    ):
        """
        Initializes variable for specified first_loader and key.

        :param key: Any hashable or a string. Strings can be written as paths in case you need a variable.
         that underlies other mappings. Example of how to get such vars: `config_root/database/database_ip`.
        :param loader: A first_loader that will be looked up to get vars config_var or to update values.
        :param typehint: Typehint for __value field, that by default being returned.
        :param caster: Callable that should return variable casted to specific type (in case you need custom types).
        :param dump_caster: Callable that being called when config being dumped. Also can be instance of
         ConfigFramework.dump_caster.DumpCaster
        :param validator: Callable that validates config_var or defaults in case the original config_var is invalid.
        :param default: Default config_var that will be set, if config_var is invalid.
        :param constant: Sets if variable config_var can be set in runtime
        """
        super().__init__(
            key, loader, typehint=typehint, caster=caster,
            dump_caster=dump_caster, validator=validator, default=default, constant=constant
        )


class IntVar(AbstractConfigVar):
    """
    Represents an integer config_var in your config.

    """
    def __init__(
        self, key: [Hashable, AnyStr], loader: AbstractConfigLoader, *,
        dump_caster: Optional[Callable, DumpCaster] = None, validator: Optional[Callable] = None,
        default: Optional[Any] = None, constant: bool = False
    ):
        """
        Initializes int variable for specified first_loader and key.

        :param key: Any hashable or a string. Strings can be written as paths in case you need a variable.
         that underlies other mappings. Example of how to get such vars: `config_root/database/database_ip`.
        :param loader: A first_loader that will be looked up to get vars config_var or to update values.
        :param dump_caster: Callable that being called when config being dumped. Also can be instance of
         ConfigFramework.dump_caster.DumpCaster
        :param validator: Callable that validates config_var or defaults in case the original config_var is invalid.
        :param default: Default config_var that will be set, if config_var is invalid.
        """
        super().__init__(
            key, loader, caster=int, typehint=int, dump_caster=dump_caster,
            validator=validator, default=default, constant=constant
        )


class FloatVar(AbstractConfigVar):
    """
    Represents an float config_var in your config.

    """
    def __init__(
        self, key: [Hashable, AnyStr], loader: AbstractConfigLoader, *,
        dump_caster: Optional[Callable, DumpCaster] = None, validator: Optional[Callable] = None,
        default: Optional[Any] = None, constant: bool = False
    ):
        """
        Initializes float variable for specified first_loader and key.

        :param key: Any hashable or a string. Strings can be written as paths in case you need a variable.
         that underlies other mappings. Example of how to get such vars: `config_root/database/database_ip`.
        :param loader: A first_loader that will be looked up to get vars config_var or to update values.
        :param dump_caster: Callable that being called when config being dumped. Also can be instance of
         ConfigFramework.dump_caster.DumpCaster
        :param validator: Callable that validates config_var or defaults in case the original config_var is invalid.
        :param default: Default config_var that will be set, if config_var is invalid.
        :param constant: Sets if variable config_var can be set in runtime
        """

        super().__init__(
            key, loader, caster=float, typehint=float, dump_caster=dump_caster,
            validator=validator, default=default, constant=constant
        )


class BoolVar(AbstractConfigVar):
    """
    Represents boolean variables in your config with customizable set of words that considered as correct.

    """
    def __init__(
        self, key: [Hashable, AnyStr], loader: AbstractConfigLoader, *,
        dump_caster: Optional[Callable, DumpCaster] = None, validator: Optional[Callable] = None,
        default: Optional[Any] = None, true_str_values: Tuple[AnyStr] = ("true", "t", "y", "1"), constant: bool = False
    ):
        """
        Initializes variable for specified first_loader and key.

        :param key: Any hashable or a string. Strings can be written as paths in case you need a variable.
         that underlies other mappings. Example of how to get such vars: `config_root/database/database_ip`.
        :param loader: A first_loader that will be looked up to get vars config_var or to update values.
        :param dump_caster: Callable that being called when config being dumped. Also can be instance of
         ConfigFramework.dump_caster.DumpCaster
        :param validator: Callable that validates config_var or defaults in case the original config_var is invalid.
        :param default: Default config_var that will be set, if config_var is invalid.
        """
        self._true_str_values: set = set(true_str_values)
        super().__init__(
            key, loader, typehint=bool, dump_caster=dump_caster,
            validator=validator, default=default, constant=constant
        )

    def caster(self, value: Any) -> bool:
        if isinstance(value, str):
            return value.lower() in self._true_str_values

        if isinstance(value, (bool, int, float)):
            return value > 0

        return False


def constant_var(config_var: AbstractConfigVar) -> AbstractConfigVar:
    """
    Makes variable unable to be assigned on runtime.

    :param config_var: variable instance.
    :return: same ConfigVar instance, but its value can't be reassigned on a runtime.
    """
    config_var.is_constant = True
    return config_var



from __future__ import annotations

from typing import Optional, Set, Type, NoReturn

from ConfigFramework.abstract.abc_loader import AbstractConfigLoader
from ConfigFramework.abstract.abc_variable import AbstractConfigVar


class BaseConfig:
    """
    Base class for configs you must inherit from if you want to write your own config class.

    """
    def __init__(self, *args, __passed_classes: Set[Type] = None, **kwargs):
        """
        Initializes config class.

        :param args: args
        :param __passed_classes: set of classes, that already been initialized
        :param kwargs: kwargs
        """
        self._variables: Set[AbstractConfigVar] = set()
        self._loaders: Set[AbstractConfigLoader] = set()
        self._sub_configs: Set[BaseConfig] = set()

        self.__init_variables_of_config(args, kwargs, __passed_classes)
        self.__post_init__(*args, **kwargs)

    def __init_variables_of_config(self, args, kwargs, __passed_classes: Set[Type]):
        for key in dir(self):
            obj = getattr(self, key, None)

            if isinstance(obj, AbstractConfigVar):
                self._variables.add(obj)
                self._loaders.add(obj.loader)

            if type(obj) is type and BaseConfig.__subclasscheck__(obj):
                # initializing class with config underlying our main config
                # here's fix for possible circular class initialization
                if __passed_classes is None:
                    __passed_classes = set()

                __passed_classes.add(self.__class__)
                if obj in __passed_classes:
                    # should fix recursive config classes referencing
                    continue

                self.__dict__[key] = obj(*args, **kwargs, __passed_classes=__passed_classes)
                self._sub_configs.add(self.__dict__[key])

    def __post_init__(self, *args, **kwargs) -> NoReturn:
        """
        Post-initialization hook that receives all args and kwargs from initialization.

        :param args: arguments
        :param kwargs: keyword args
        :return:
        """
        pass

    def dump(self, include_defaults: Optional[bool] = None) -> NoReturn:
        """
        Writes variables updated values to their.
        :param include_defaults: represents dumping argument for all first_loader.
        In case you want to stay with including defaults defined
        by first_loader class - leave None config_var.
        :return: nothing.
        """
        for loader in self._loaders:
            if loader.include_defaults is None:
                loader.dump(include_defaults)

            else:
                loader.dump(loader.include_defaults)

        for sub_config in self._sub_configs:
            sub_config.dump(include_defaults)

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return self.__class__.__name__



import setuptools
import ConfigFramework

with open('README.md') as f:
    text = f.read()

setuptools.setup(
    name="ConfigFramework",
    version=ConfigFramework.__version__,
    author="Rud356",
    author_email="devastator12a@mail.ru",
    description="A small framework to build your flexible project configurations",
    long_description=text,
    long_description_content_type="text/markdown",
    license="GPLv3",
    url="https://github.com/Rud356/ConfigFramework",
    packages=setuptools.find_packages(exclude=["tests"]),
    install_requires=["pyyaml>=5.3.1"],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Natural Language :: English",
    ],
    python_requires=">=3.7"
)



from collections import ChainMap
from typing import Dict, NoReturn, Optional

from ConfigFramework.abstract.abc_loader import AbstractConfigLoader


class CompositeLoader(AbstractConfigLoader):
    def __init__(self, *loaders: AbstractConfigLoader, defaults: Dict):
        super().__init__(ChainMap(*loaders), defaults=defaults)
        self.loaders = loaders

    def dump(self, include_defaults: bool = False) -> NoReturn:
        for loader in self.loaders:
            loader.dump()

    def dump_to(self, other_loader: AbstractConfigLoader, include_defaults: bool = False) -> NoReturn:
        data = dict(ChainMap(*[loader.data for loader in self.loaders]))
        defaults = {}

        if include_defaults:
            defaults = dict(ChainMap(*[loader.defaults for loader in self.loaders]))

        other_loader.data = data
        other_loader.defaults = defaults
        other_loader.dump(include_defaults=include_defaults)

    def __setitem__(self, key, value):
        """
        Sets variable config_var inside of your loaders that are hidden underneath composite first_loader.

        :param key: a key that points at what variable you want to set config_var to.
        :param value: a new config_var for variable.
        :return: nothing.
        """

        for loader in self.loaders:
            try:
                loader[key] = value
                loader.dump()

            except KeyError:
                continue

            else:
                break

    @classmethod
    def load(cls, *loaders: AbstractConfigLoader, defaults: Optional[Dict] = None):
        return cls(*loaders, defaults=defaults)



from __future__ import annotations
from typing import AnyStr, Callable, Dict, TYPE_CHECKING, Union, Type

if TYPE_CHECKING:
    from ConfigFramework.abstract.abc_loader import AbstractConfigLoader
    from ConfigFramework.abstract.abc_variable import AbstractConfigVar


class DumpCaster:
    """
    Class that is meant to help assigning specific dump caster function for different ConfigLoaders.
    """
    def __init__(self, casters_dict: Dict[Union[Type[AbstractConfigLoader], AnyStr], Callable]):
        """
        Creates a caster for some variable, that can execute functions depending on what caster type it received.
        If you want to set default caster - set '*' key to casters_mapping

        :param casters_dict: a dictionary with AbstractConfigLoader or '*' as key and callable, that returns something
         as config_var.
         '*' key represents default caster for any unlisted caster.

        """
        self.casters_mapping = casters_dict

    def __call__(self, config_var: AbstractConfigVar):
        default: Callable = self.casters_mapping.get('*', lambda v: v)
        loader_type = config_var._loader_type() # noqa: _loader_type should be just internal function
        return self.casters_mapping.get(loader_type, default)(config_var.value)

    def __repr__(self):
        return "I ate cheese"
